// Fill out your copyright notice in the Description page of Project Settings.

#include "TerrainChunk.h"
#include "ProceduralMeshComponent.h"
#include "KismetProceduralMeshLibrary.h"
#include "MeshDescription.h"

constexpr int16 G_Max_Depth = 4;

ATerrainChunk::ATerrainChunk()
{
	PrimaryActorTick.bCanEverTick = true;
	
	ProcMeshComp = CreateDefaultSubobject<UProceduralMeshComponent>(TEXT("Procedural Mesh"));
	RootComponent = ProcMeshComp;
}

// Called when the game starts or when spawned
void ATerrainChunk::BeginPlay()
{
	Super::BeginPlay();
}

// Called every frame
void ATerrainChunk::Tick(float DeltaTime)
{
	return;
	Super::Tick(DeltaTime);
	
	if (Planet && Planet->ChunkDistanceActor) {
		const float Distance = FVector::Distance(Planet->ChunkDistanceActor->GetActorLocation(), GetActorLocation());
		 
		if (Depth < Planet->ChunkLodLevels && Distance < DistanceToSplit) {
			Split();
		} else if (Depth > 0 && Distance > DistanceToJoin) {
			UnSplit();
		}
	}
}

void ATerrainChunk::InitQuadTree(AProceduralPlanet* PlanetActor, int32 ChunkDepth)
{
	Planet = PlanetActor;
	Depth = ChunkDepth;
}

void ATerrainChunk::InitMesh(int32 Resolution, FVector Offset, float Radius)
{
	ChunkResolution = Resolution;
	ChunkRadius = Radius;
	DistanceToSplit = ChunkRadius * 1.25;
	DistanceToJoin = ChunkRadius * 1.5;
	
	TArray<FVector> Vertices;
	TArray<FVector> Normals;
	TArray<int32> Triangles;
	TArray<FVector2D> UVs;
	TArray<FLinearColor> VertexColors;
	TArray<FProcMeshTangent> Tangents;
	
	UKismetProceduralMeshLibrary::CreateGridMeshWelded(Resolution + 1, Resolution + 1, Triangles, Vertices, UVs, ChunkRadius / Resolution);

	for (FVector &Vertex : Vertices) {
		Vertex += Offset;
	}

	ProcMeshComp->CreateMeshSection_LinearColor(0, Vertices, Triangles, Normals, UVs, VertexColors, Tangents, true);
}

void ATerrainChunk::Update()
{
	ProcMeshComp->SetMeshSectionVisible(0, true);
	ProcMeshComp->RegisterComponent();
	//ProcMeshComp->SetProcMeshSection(0, *ProcMeshComp->GetProcMeshSection(0));
	return;
	
	TArray<FVector> Vertices;
	TArray<FVector> Normals;
	TArray<int32> Triangles;
	TArray<FVector2D> UVs;
	TArray<FLinearColor> VertexColors;
	TArray<FProcMeshTangent> Tangents;
	
	UKismetProceduralMeshLibrary::CreateGridMeshWelded(ChunkResolution + 1, ChunkResolution + 1, Triangles, Vertices, UVs, ChunkRadius / ChunkResolution);
	
	ProcMeshComp->UpdateMeshSection_LinearColor(0, Vertices, Normals, UVs, UVs, UVs, UVs, VertexColors, Tangents);
}

void ATerrainChunk::Split()
{
	ProcMeshComp->ClearAllMeshSections();
	Destroy();
	
	// create child with depth + 1
	
	const float HalfRadius = ChunkRadius / 2.0;
	const float QuarterRadius = ChunkRadius / 4.0;

	ATerrainChunk* Chunk1 = GetWorld()->SpawnActor<ATerrainChunk>();
	Chunk1->InitQuadTree(Planet, Depth + 1);
	Chunk1->InitMesh(ChunkResolution * Planet->ChunkResolutionIncreasePerLodLevel, FVector(-QuarterRadius,-QuarterRadius,0), HalfRadius);

	ATerrainChunk* Chunk2 = GetWorld()->SpawnActor<ATerrainChunk>();
	Chunk2->InitQuadTree(Planet, Depth + 1);
	Chunk2->InitMesh(ChunkResolution * Planet->ChunkResolutionIncreasePerLodLevel, FVector(-QuarterRadius,QuarterRadius,0), HalfRadius);

	ATerrainChunk* Chunk3 = GetWorld()->SpawnActor<ATerrainChunk>();
	Chunk3->InitQuadTree(Planet, Depth + 1);
	Chunk3->InitMesh(ChunkResolution * Planet->ChunkResolutionIncreasePerLodLevel, FVector(QuarterRadius,-QuarterRadius,0), HalfRadius);

	ATerrainChunk* Chunk4 = GetWorld()->SpawnActor<ATerrainChunk>();
	Chunk4->InitQuadTree(Planet, Depth + 1);
	Chunk4->InitMesh(ChunkResolution * Planet->ChunkResolutionIncreasePerLodLevel, FVector(QuarterRadius,QuarterRadius,0), HalfRadius);
}

void ATerrainChunk::UnSplit()
{
	// tell parent. Parent will destroy children and recreate itself
	// - or parent waits until all children can be lower detail and then?
	//Destroy();
}

